
exception Unimplemented
exception RuntimeTypeError
exception DoesNotTypecheck of string

(****** Syntax for our language, including types (do not change) *****)
type exp = Var of string 
         | Lam of string * typ * exp
	 | Apply of exp * exp
	 | Closure of string * exp * (env ref)
	 | Int of int
	 | Plus of exp * exp
	 | If of exp * exp * exp
	 | RecordE of (string * exp) list
	 | RecordV of (string * (exp ref)) list (* all exps are value *)
	 | Get of exp * string
	 | Set of exp * string * exp
	 | Letrec of typ * string *  string * typ * exp
	 | Cast of exp * typ

and env = (string * exp) list

and access = Read | Write | Both

and typ = IntT
        | ArrowT of typ * typ
	| RecordT of (string * typ * access) list

and ctxt = (string * typ) list

(****** Interpreter for our language (do not change) *****)
let rec interp env e =
  match e with
    Var s -> (try List.assoc s env with Not_found -> raise RuntimeTypeError)
  | Lam(s,_,e2) -> Closure(s,e2,ref env) 
  | Closure _ -> e 
  | Apply(e1,e2) ->
      let v1 = interp env e1 in
      let v2 = interp env e2 in
      (match v1 with
	Closure(s,e3,env2) -> interp((s,v2)::(!env2)) e3
      | _ -> raise RuntimeTypeError)
  | Plus(e1,e2) ->
      let v1 = interp env e1 in
      let v2 = interp env e2 in
      (match(v1,v2) with
	(Int i, Int j) -> Int (i+j)
      | _ -> raise RuntimeTypeError)
  | If(e1,e2,e3) ->
      let v1 = interp env e1 in
      (match v1 with
	Int 0 -> interp env e3
      | Int _ -> interp env e2
      | _ -> raise RuntimeTypeError)
  | Int _ -> e
  | RecordV _ -> e
  | RecordE lst -> RecordV (List.map (fun (s,r) -> (s,ref(interp env r))) lst)
  | Get(e,s) -> 
      (match interp env e with
	RecordV lst -> 
	  (try !(List.assoc s lst) with Not_found -> raise RuntimeTypeError)
      | _ -> raise RuntimeTypeError)
  | Set(e1,s,e2) ->
      (match interp env e1 with	
	RecordV lst -> 
	  let r=try List.assoc s lst with Not_found -> raise RuntimeTypeError in
	  let ans = interp env e2 in 
	  r := ans; ans
      | _ -> raise RuntimeTypeError)
  | Letrec(_,f,x,_,e) ->
      let r = ref env in
      let c = Closure(x,e,r) in
      let _ = r := (f,c)::(!r) in
      c
  | Cast(e,t) -> interp env e

let interp e = interp [] e

(***** helper functions provided to you (do not change) *****)
let fields_unique lst = (*raise exception if same field name appears > 1 time*)
  let rec loop lst1 lst2 =
    match lst2 with
      [] -> ()
    | (s,t,a)::tl -> 
	if (List.exists (fun (s2,_,_) -> s2=s) lst1)
	then raise (DoesNotTypecheck "")
	else loop ((s,t,a)::lst1) tl in
  loop [] lst

let rec checkType t = (* checks field_unique for every contained record-type *)
  match t with
    IntT -> ()
  | ArrowT(t1,t2) -> checkType t1; checkType t2
  | RecordT lst -> fields_unique lst; List.iter (fun (_,t,_) -> checkType t) lst

let rec getFieldType lst str = (* (string * typ * access) list -> string -> (typ * access) option *)
  match lst with
    [] -> None
  | (s,t,a)::tl -> if s=str then Some (t,a) else getFieldType tl str

(******** Problem 1: complete subtype and typecheck *********)
let rec subtype t1 t2 = (* true if t1 is a subtype of t2 *)
  match t1,t2 with
    _ -> raise Unimplemented

let rec typecheck ctxt e = (* return type of e under ctxt *)
  (* we did 5 cases for you *)
  match e with
    Var s -> (try List.assoc s ctxt with Not_found -> raise (DoesNotTypecheck ""))
  | Lam(s,t,e) -> checkType t; ArrowT(t,typecheck ((s,t)::ctxt) e)
  | Closure _ -> raise (DoesNotTypecheck "not a source program")
  | Int _ -> IntT
  | Plus(e1,e2) ->
      if subtype (typecheck ctxt e1) IntT && subtype (typecheck ctxt e2) IntT
      then IntT
      else raise (DoesNotTypecheck "")
  | _ -> raise Unimplemented

let typecheck e = typecheck [] e

(********** examples and testing ***********)

(* we have provided two tests that should type-check, but you can make small
   changes to them that should cause them not to type-check *)

(* first some helper metafunctions to make examples easier to write *)
let lam x t e = Lam(x,t,e)
let app e1 e2 = Apply(e1,e2)
let vx = Var "x"
let vy = Var "y"
let vf = Var "f"

(* first test -- puts sum of two fields in second field *)
(* (\ x : {l1=int,read, l2=int,both}. x.l2 := x.l1 + x.l2) {l3=12,l2=17,l1=19}*)
let ex1 = 
  app (lam "x" (RecordT [("l1",IntT,Read);("l2",IntT,Both)])
	 (Set(vx,"l2",Plus(Get(vx,"l1"),Get(vx,"l2")))))
      (RecordE [("l3",Int 12);("l2",Int 17);("l1",Int 19)])

let t1 = typecheck ex1
let v1 = interp ex1

(* second test -- uses letrec to implement multiplication of positive nums *)
let ex2 = 
  app (Letrec(IntT,"f","x",(RecordT [("l1",IntT,Read);("l2",IntT,Read)]),
	app (lam "minus1" IntT
	(If(Var "minus1",
            Plus(Get(vx,"l2"),
                 app vf (RecordE[("l1", Var "minus1");
                                  ("l2", Get(vx,"l2"))])),
            Get(vx,"l2"))))
        (Plus(Get(vx,"l1"),Int(-1)))))
   (RecordE[("l1", Int 7); ("l2", Int 6)])

let t2 = typecheck ex2
let v2 = interp ex2

(* for problem 2 *)
let t3 = typecheck (Lam("x",RecordT[("l",IntT,Read)],
			RecordE[("l1",Get(Var("x"),"l")); ("l2",Var("x"))]))
let t4 = typecheck (Lam("x",RecordT[("l",IntT,Both)],
			RecordE[("l1",Get(Var("x"),"l")); ("l2",Var("x"))]))

